@@ -1,5 +1,5 @@
 /*-
- * Copyright (c) 2009-2010 EPIPE Communications. All rights reserved.
+ * Copyright (c) 2009-2011 EPIPE Communications. All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -48,17 +48,28 @@
  * on GNU/Linux.
  */
 
-#ifdef __linux__
+#if defined(__linux__) || defined(__GLIBC__)
+/* GNU/Linux or GNU/kFreeBSD */
 #define	_GNU_SOURCE
 #define	_BSD_SOURCE
 #endif
 
+/*
+ * Pull in facilitynames array if <syslog.h> has it.
+ */
+
+#ifdef HAVE_SYSLOG_FACILITYNAMES
+#define	SYSLOG_NAMES
+#endif
+
 /* Standard includes that we need */
 
 #include <sys/file.h>
 #include <sys/stat.h>
-#ifdef HAVE_SYS_SYSCTL_H
+#ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
+#endif
+#ifdef HAVE_SYS_SYSCTL_H
 #include <sys/sysctl.h>
 #endif
 #include <sys/time.h>
@@ -68,7 +79,9 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <inttypes.h>
+#ifdef HAVE_PATHS_H
 #include <paths.h>
+#endif
 #include <signal.h>
 #include <stdarg.h>
 #include <stdio.h>
@@ -77,8 +90,9 @@
 #include <sysexits.h>
 #include <time.h>
 #include <unistd.h>
-
-#define  SYSLOG_NAMES
+#if defined(__SVR4) && defined(HAVE_UTMPX_H)
+#include <utmpx.h>
+#endif
 #include <syslog.h>
 
 #include "downtimedb.h"
@@ -99,14 +113,17 @@
 /* from <paths.h> */
 
 #ifndef _PATH_VARRUN
-#error _PATH_VARRUN is not defined
+#define	_PATH_VARRUN	"/var/run/"
+#endif
+
+#ifndef _PATH_DEVNULL
+#define	_PATH_DEVNULL	"/dev/null"
 #endif
 
 /* from <sys/stat.h> */
 
 #ifndef DEFFILEMODE
 #define	DEFFILEMODE 0666
-#warn DEFFILEMODE was not defined
 #endif
 
 /* Function prototypes */
@@ -126,13 +143,19 @@ static void	parseargs(int, char *[]);
 static int	makepidfile(void);
 static void	removepidfile(void);
 
+#ifndef HAVE_DAEMON
+static int	daemon(int, int);
+#endif
+
 /* Command line arguments with their defaults */
 
 char *	cf_log = "daemon";  /* syslog facility or filename with a slash (/) */
 char *	cf_pidfile = _PATH_VARRUN PROGNAME ".pid";
 char *	cf_datadir = PATH_DOWNTIMEDBDIR;
 long	cf_sleep = 15;                 /* update time stamp every 15 seconds */
+#ifdef HAVE_FUTIMES
 int	cf_fsync = 1;	      /* if true, fsync() stamp files after touching */
+#endif
 int	cf_downtimedb = 1;		       /* if true, update downtimedb */
 char *	cf_downtimedbfile = PATH_DOWNTIMEDBFILE;
 
@@ -143,14 +166,17 @@ int	cf_logfd = -1;
 
 /* Global variables */
 
-int	exiting = 0;
-int	reopenlog = 0;
 char *	ts_stamp = NULL;
 char *	ts_shutdown = NULL;
 char *	ts_boot = NULL;
 time_t	boottime = 0;
 time_t	starttime = 0;
 
+/* The following are set by the signal handler */
+
+volatile sig_atomic_t	exiting = 0;
+volatile sig_atomic_t	reopenlog = 0;
+
 /*
  * downtimed: system downtime monitoring and reporting daemon.
  *
@@ -289,7 +315,7 @@ getboottime()
 	 */
 	char str[1024];
 	FILE *fp;
-	time_t bt = 0;
+	long unsigned bt = 0;
 
 	if ((fp = fopen("/proc/stat", "r")) != NULL) {
 		while (fgets(str, sizeof(str), fp) != NULL)
@@ -297,10 +323,28 @@ getboottime()
 				if (sscanf(str, "btime %lu", &bt) == 1 &&
 				    bt != 0) {
 					fclose(fp);
-					return (bt);
+					return ((time_t)bt);
 				}
 		fclose(fp);
 	}
+#elif defined(__SVR4) && defined(HAVE_UTMPX_H)
+	/*
+	 * This should work on some SVR4 systems, such as Solaris
+	 * and possibly HP-UX.
+	 */
+	struct utmpx ut, *utp;
+
+	memset(&ut, 0, sizeof(ut));
+	ut.ut_type = BOOT_TIME;
+	if ((utp = getutxid(&ut)) != NULL) {
+		time_t boottime;
+
+		boottime = utp->ut_tv.tv_sec;
+		endutxent();
+
+		return (boottime);
+	}
+	endutxent();
 #endif
 	/*
 	 * We fall through here if we do not have OS specific code
@@ -311,6 +355,8 @@ getboottime()
 	 * system log, dmesg.boot, kernel, utmp, etc.) XXX
 	 */
 
+	logwr(LOG_ERR, "can not determine system boot time on this OS");
+
 	return (starttime);	/* give up */
 }
 
@@ -445,6 +491,7 @@ sighandler(int signum)
 static void
 touch(const char *fn, time_t t)
 {
+	struct stat sb;
 	struct timeval tv[2];
 	int fd;
 
@@ -455,6 +502,7 @@ touch(const char *fn, time_t t)
 		tv[1].tv_usec = 0;
 	}
 
+#ifdef HAVE_FUTIMES
 	if (cf_fsync) {
 		/* we need to open the file so that we can do fsync() to it */
 		if ((fd = open(fn, O_WRONLY | O_CREAT | O_TRUNC,
@@ -470,12 +518,92 @@ touch(const char *fn, time_t t)
 		if (close(fd) < 0)
 			logwr(LOG_ERR, "%s: %s", fn, strerror(errno));
 	} else {
-		/* not doing fsync(), no need to open file */
+#endif /* HAVE_FUTIMES */
+		/* create the file in case it is missing */
+		if (stat(fn, &sb) < 0) {
+			if ((fd = open(fn, O_WRONLY | O_CREAT | O_TRUNC,
+			    DEFFILEMODE)) < 0) {
+				logwr(LOG_ERR, "%s: %s", fn, strerror(errno));
+				return;
+			}
+			if (close(fd) < 0)
+				logwr(LOG_ERR, "%s: %s", fn, strerror(errno));
+		}
 		if (utimes(fn, t == 0 ? (struct timeval *)NULL : tv) < 0)
 			logwr(LOG_ERR, "%s: %s", fn, strerror(errno));
+#ifdef HAVE_FUTIMES
 	}
+#endif /* HAVE_FUTIMES */
 }
 
+/* Compatibility for systems without facilitynames in <syslog.h> */
+
+#ifndef HAVE_SYSLOG_FACILITYNAMES
+
+static const struct {
+	const char	*c_name;
+	const int	c_val;
+} facilitynames[] = {
+#ifdef LOG_KERN
+	{ "kern",	LOG_KERN	},
+#endif
+#ifdef LOG_USER
+	{ "user",	LOG_USER	},
+#endif
+#ifdef LOG_MAIL
+	{ "mail",	LOG_MAIL	},
+#endif
+#ifdef LOG_DAEMON
+	{ "daemon",	LOG_DAEMON	},
+#endif
+#ifdef LOG_AUTH
+	{ "auth",	LOG_AUTH	},
+#endif
+#ifdef LOG_SYSLOG
+	{ "syslog",	LOG_SYSLOG	},
+#endif
+#ifdef LOG_LPR
+	{ "lpr",	LOG_LPR		},
+#endif
+#ifdef LOG_NEWS
+	{ "news",	LOG_NEWS	},
+#endif
+#ifdef LOG_AUDIT
+	{ "audit",	LOG_AUDIT	},
+#endif
+#ifdef LOG_CRON
+	{ "cron",	LOG_CRON	},
+#endif
+#ifdef LOG_LOCAL0
+	{ "local0",	LOG_LOCAL0	},
+#endif
+#ifdef LOG_LOCAL1
+	{ "local1",	LOG_LOCAL1	},
+#endif
+#ifdef LOG_LOCAL2
+	{ "local2",	LOG_LOCAL2	},
+#endif
+#ifdef LOG_LOCAL3
+	{ "local3",	LOG_LOCAL3	},
+#endif
+#ifdef LOG_LOCAL4
+	{ "local4",	LOG_LOCAL4	},
+#endif
+#ifdef LOG_LOCAL5
+	{ "local5",	LOG_LOCAL5	},
+#endif
+#ifdef LOG_LOCAL6
+	{ "local6",	LOG_LOCAL6	},
+#endif
+#ifdef LOG_LOCAL7
+	{ "local7",	LOG_LOCAL7	},
+#endif
+	{ NULL,		-1		}
+};
+
+
+#endif /* !defined(HAVE_SYSLOG_FACILITYNAMES) */
+
 /* Determine log destination & initialize */
 
 static void
@@ -578,7 +706,7 @@ version()
 
 	puts(PROGNAME " " PROGVERSION " - system downtime reporting daemon\n");
 
-	puts("Copyright (c) 2009-2010 EPIPE Communications. "
+	puts("Copyright (c) 2009-2011 EPIPE Communications. "
 	    "All rights reserved.");
 	puts("This software is licensed under the terms and conditions of the "
 	    "FreeBSD");
@@ -592,7 +720,9 @@ version()
 	printf("  datadir = %s\n", cf_datadir);
 	printf("  downtimedbfile = %s\n", cf_downtimedbfile);
 	printf("  sleep = %ld\n", cf_sleep);
+#ifdef HAVE_FUTIMES
 	printf("  fsync = %d\n", cf_fsync);
+#endif
 
 #ifdef PACKAGE_URL
 	puts("\nSee the following web site for more information and updates:");
@@ -630,9 +760,11 @@ parseargs(int argc, char *argv[])
 			if ((p != NULL && *p != '\0') || errno != 0)
 				errx(EX_USAGE, "-s argument is not a number");
 			break;
+#ifdef HAVE_FUTIMES
 		case 'S':
 			cf_fsync = 0;
 			break;
+#endif
 		case 'v':
 			version();
 			/* NOTREACHED */
@@ -667,11 +799,15 @@ retry:
 		    cf_pidfile, strerror(errno));
 		return (-1);
 	}
+#ifdef HAVE_FLOCK
 	if (flock(fd, LOCK_EX | LOCK_NB) < 0) {
-		if (errno == EWOULDBLOCK)
+#else
+	if (lockf(fd, F_TLOCK, 0) < 0) {
+#endif
+		if (errno == EWOULDBLOCK || errno == EAGAIN || errno == EACCES)
 			logwr(LOG_ERR, "another process is running already");
 		else
-			logwr(LOG_ERR, "can not flock pid file %s: %s",
+			logwr(LOG_ERR, "can not lock pid file %s: %s",
 			    cf_pidfile, strerror(errno));
 		close(fd);
 		return (-1);
@@ -700,7 +836,7 @@ retry:
 		close(fd);
 		return (-1);
 	}
-	snprintf(str, sizeof(str), "%d\n", getpid());
+	snprintf(str, sizeof(str), "%ld\n", (long)getpid());
 	if (write(fd, str, strlen(str)) != strlen(str)) {
 		logwr(LOG_ERR, "can not write pid file %s: %s",
 		    cf_pidfile, strerror(errno));
@@ -732,4 +868,38 @@ removepidfile()
 	unlink(cf_pidfile);
 }
 
+/* Compatibility daemon(3) function for OSes which lack it */
+
+#ifndef HAVE_DAEMON
+static int
+daemon(int nochdir, int noclose)
+{
+	int fd;
+
+	switch (fork()) {
+	case -1:
+		return (-1);
+	case 0:
+		break;
+	default:
+		exit(0);
+	}
+
+	if (setsid() == -1)
+		return (-1);
+
+	if (!nochdir)
+		chdir("/");
+
+	if (!noclose && (fd = open(_PATH_DEVNULL, O_RDWR, 0)) != -1) {
+		dup2(fd, STDIN_FILENO);
+		dup2(fd, STDOUT_FILENO);
+		dup2(fd, STDERR_FILENO);
+		if (fd > STDERR_FILENO)
+			close(fd);
+	}
+	return (0);
+}
+#endif /* !HAVE_DAEMON */
+
 /* eof */
